# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List
from unittest import TestCase

import pytest
from license_expression import LicenseExpression, get_spdx_licensing

from spdx_tools.spdx.model import Document, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.validation.license_expression_validator import (
    validate_license_expression,
    validate_license_expressions,
)
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import document_fixture, extracted_licensing_info_fixture

FIXTURE_LICENSE_ID = extracted_licensing_info_fixture().license_id


@pytest.mark.parametrize(
    "expression_string",
    [
        "MIT",
        FIXTURE_LICENSE_ID,
        f"GPL-2.0-only with GPL-CC-1.0 and {FIXTURE_LICENSE_ID} with 389-exception or Beerware",
    ],
)
def test_valid_license_expression(expression_string):
    document: Document = document_fixture()
    license_expression: LicenseExpression = get_spdx_licensing().parse(expression_string)
    validation_messages: List[ValidationMessage] = validate_license_expression(
        license_expression, document, parent_id="SPDXRef-File"
    )

    assert validation_messages == []


@pytest.mark.parametrize("expression", [SpdxNone(), SpdxNoAssertion()])
def test_none_and_no_assertion(expression):
    document: Document = document_fixture()
    validation_messages: List[ValidationMessage] = validate_license_expression(
        expression, document, parent_id="SPDXRef-File"
    )
    assert validation_messages == []


@pytest.mark.parametrize(
    "expression_list",
    [
        [SpdxNone()],
        [SpdxNoAssertion()],
        [get_spdx_licensing().parse("MIT and GPL-3.0-only"), get_spdx_licensing().parse(FIXTURE_LICENSE_ID)],
        [SpdxNone(), get_spdx_licensing().parse("MIT"), SpdxNoAssertion()],
    ],
)
def test_valid_license_expressions(expression_list):
    document: Document = document_fixture()
    validation_messages: List[ValidationMessage] = validate_license_expressions(
        expression_list, document, parent_id="SPDXRef-File"
    )
    assert validation_messages == []


@pytest.mark.parametrize(
    "expression_string, unknown_symbols",
    [
        (f"{FIXTURE_LICENSE_ID} or LicenseRef-22", ["LicenseRef-22"]),
        ("nope with 389-exception and _.- or LicenseRef-10", ["nope", "_.-", "LicenseRef-10"]),
    ],
)
def test_invalid_license_expression_with_unknown_symbols(expression_string, unknown_symbols):
    document: Document = document_fixture()
    license_expression: LicenseExpression = get_spdx_licensing().parse(expression_string)
    parent_id = "SPDXRef-File"
    context = ValidationContext(
        parent_id=parent_id, element_type=SpdxElementType.LICENSE_EXPRESSION, full_element=license_expression
    )

    validation_messages: List[ValidationMessage] = validate_license_expression(license_expression, document, parent_id)
    expected_messages = [
        ValidationMessage(
            f"Unrecognized license reference: {symbol}. license_expression must only use IDs from the license list or "
            f"extracted licensing info, but is: {license_expression}",
            context,
        )
        for symbol in unknown_symbols
    ]

    TestCase().assertCountEqual(validation_messages, expected_messages)


@pytest.mark.parametrize(
    "expression_string, expected_message",
    [
        (
            "MIT with MIT",
            'A plain license symbol cannot be used as an exception in a "WITH symbol" statement. for token: "MIT" at '
            "position: 9. for license_expression: MIT WITH MIT",
        ),
        (
            f"GPL-2.0-or-later and {FIXTURE_LICENSE_ID} with {FIXTURE_LICENSE_ID}",
            f'A plain license symbol cannot be used as an exception in a "WITH symbol" statement. for token: '
            f'"{FIXTURE_LICENSE_ID}" at position: 39. for license_expression: GPL-2.0-or-later AND '
            f"{FIXTURE_LICENSE_ID} WITH {FIXTURE_LICENSE_ID}",
        ),
        (
            f"GPL-2.0-or-later with MIT and {FIXTURE_LICENSE_ID} with GPL-2.0-or-later",
            f'A plain license symbol cannot be used as an exception in a "WITH symbol" statement. for token: "MIT" at '
            f"position: 22. for license_expression: GPL-2.0-or-later WITH MIT AND {FIXTURE_LICENSE_ID} "
            f"WITH GPL-2.0-or-later",
        ),
        (
            "389-exception with 389-exception",
            'A license exception symbol can only be used as an exception in a "WITH exception" statement. for token: '
            '"389-exception". for license_expression: 389-exception WITH 389-exception',
        ),
        (
            "389-exception with MIT",
            'A license exception symbol can only be used as an exception in a "WITH exception" statement. for token: '
            '"389-exception". for license_expression: 389-exception WITH MIT',
        ),
    ],
)
def test_invalid_license_expression_with_invalid_exceptions(expression_string, expected_message):
    document: Document = document_fixture()
    license_expression: LicenseExpression = get_spdx_licensing().parse(expression_string)
    parent_id = "SPDXRef-File"
    context = ValidationContext(
        parent_id=parent_id, element_type=SpdxElementType.LICENSE_EXPRESSION, full_element=license_expression
    )

    validation_messages: List[ValidationMessage] = validate_license_expression(license_expression, document, parent_id)
    expected_messages = [ValidationMessage(expected_message, context)]

    assert validation_messages == expected_messages



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDF, RDFS, Graph, Literal, URIRef

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.rdfschema.namespace import LICENSE_NAMESPACE, SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.creation_info_writer import add_creation_info_to_graph
from tests.spdx.fixtures import creation_info_fixture


def test_add_creation_info_to_graph():
    graph = Graph()
    creation_info = creation_info_fixture()

    add_creation_info_to_graph(creation_info, graph)

    assert (None, RDF.type, SPDX_NAMESPACE.SpdxDocument) in graph
    assert (URIRef(f"{creation_info.document_namespace}#{creation_info.spdx_id}"), None, None) in graph
    assert (None, SPDX_NAMESPACE.dataLicense, LICENSE_NAMESPACE[creation_info.data_license]) in graph
    assert (None, SPDX_NAMESPACE.name, Literal(creation_info.name)) in graph
    assert (None, SPDX_NAMESPACE.specVersion, Literal(creation_info.spdx_version)) in graph
    assert (None, SPDX_NAMESPACE.creationInfo, None) in graph

    assert (None, RDF.type, SPDX_NAMESPACE.CreationInfo) in graph
    assert (None, SPDX_NAMESPACE.created, Literal(datetime_to_iso_string(creation_info.created))) in graph
    assert (None, RDFS.comment, Literal(creation_info.creator_comment)) in graph
    assert (None, SPDX_NAMESPACE.licenseListVersion, Literal(creation_info.license_list_version)) in graph
    assert (None, SPDX_NAMESPACE.creator, Literal(creation_info.creators[0].to_serialized_string())) in graph



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import mock
from unittest.mock import MagicMock

import pytest

from spdx_tools.spdx.jsonschema.external_document_ref_converter import ExternalDocumentRefConverter
from spdx_tools.spdx.jsonschema.external_document_ref_properties import ExternalDocumentRefProperty
from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm, ExternalDocumentRef


@pytest.fixture
@mock.patch("spdx_tools.spdx.jsonschema.checksum_converter.ChecksumConverter", autospec=True)
def converter(checksum_converter_magic_mock: MagicMock) -> ExternalDocumentRefConverter:
    mocked_checksum_converter = checksum_converter_magic_mock()
    converter = ExternalDocumentRefConverter()
    converter.checksum_converter = mocked_checksum_converter
    return converter


@pytest.mark.parametrize(
    "external_document_ref_property,expected",
    [
        (ExternalDocumentRefProperty.EXTERNAL_DOCUMENT_ID, "externalDocumentId"),
        (ExternalDocumentRefProperty.SPDX_DOCUMENT, "spdxDocument"),
        (ExternalDocumentRefProperty.CHECKSUM, "checksum"),
    ],
)
def test_json_property_names(
    converter: ExternalDocumentRefConverter, external_document_ref_property: ExternalDocumentRefProperty, expected: str
):
    assert converter.json_property_name(external_document_ref_property) == expected


def test_successful_conversion(converter: ExternalDocumentRefConverter):
    converter.checksum_converter.convert.return_value = "mock_converted_checksum"
    checksum = Checksum(ChecksumAlgorithm.SHA1, "123")
    external_document_ref = ExternalDocumentRef("document_ref_id", "document_uri", checksum)

    converted_dict = converter.convert(external_document_ref)

    assert converted_dict == {
        converter.json_property_name(ExternalDocumentRefProperty.EXTERNAL_DOCUMENT_ID): "document_ref_id",
        converter.json_property_name(ExternalDocumentRefProperty.SPDX_DOCUMENT): "document_uri",
        converter.json_property_name(ExternalDocumentRefProperty.CHECKSUM): "mock_converted_checksum",
    }


def test_json_type(converter: ExternalDocumentRefConverter):
    assert converter.get_json_type() == ExternalDocumentRefProperty


def test_data_model_type(converter: ExternalDocumentRefConverter):
    assert converter.get_data_model_type() == ExternalDocumentRef



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class FileProperty(JsonProperty):
    SPDX_ID = auto()
    ANNOTATIONS = auto()
    ARTIFACT_OFS = auto()
    ATTRIBUTION_TEXTS = auto()
    CHECKSUMS = auto()
    COMMENT = auto()
    COPYRIGHT_TEXT = auto()
    FILE_CONTRIBUTORS = auto()
    FILE_DEPENDENCIES = auto()
    FILE_NAME = auto()
    FILE_TYPES = auto()
    LICENSE_COMMENTS = auto()
    LICENSE_CONCLUDED = auto()
    LICENSE_INFO_IN_FILES = auto()
    NOTICE_TEXT = auto()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

from beartype.typing import Dict, List, Optional

from spdx_tools.spdx.datetime_conversions import datetime_from_str
from spdx_tools.spdx.model import Actor, Checksum, CreationInfo, ExternalDocumentRef, Version
from spdx_tools.spdx.parser.actor_parser import ActorParser
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.checksum_parser import ChecksumParser
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import parse_field_or_log_error
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)


class CreationInfoParser:
    logger: Logger
    actor_parser: ActorParser
    checksum_parser: ChecksumParser

    def __init__(self):
        self.logger = Logger()
        self.actor_parser = ActorParser()
        self.checksum_parser = ChecksumParser()

    def parse_creation_info(self, doc_dict: Dict) -> CreationInfo:
        logger = Logger()
        spdx_version: Optional[str] = doc_dict.get("spdxVersion")
        spdx_id: Optional[str] = doc_dict.get("SPDXID")
        name: Optional[str] = doc_dict.get("name")
        document_namespace: Optional[str] = doc_dict.get("documentNamespace")
        creation_info_dict: Optional[Dict] = doc_dict.get("creationInfo")

        # There are nested required properties. If creationInfo is not set, we cannot continue parsing.
        if creation_info_dict is None:
            logger.append("CreationInfo does not exist.")
            raise SPDXParsingError([f"Error while parsing document {name}: {logger.get_messages()}"])

        creators: List[Actor] = parse_field_or_log_error(
            logger, creation_info_dict.get("creators"), self.actor_parser.parse_actor, field_is_list=True
        )

        created: Optional[datetime] = parse_field_or_log_error(
            logger, creation_info_dict.get("created"), datetime_from_str
        )

        creator_comment: Optional[str] = creation_info_dict.get("comment")
        data_license: Optional[str] = doc_dict.get("dataLicense")

        external_document_refs: List[ExternalDocumentRef] = parse_field_or_log_error(
            logger, doc_dict.get("externalDocumentRefs"), self.parse_external_document_refs
        )
        license_list_version: Optional[Version] = parse_field_or_log_error(
            logger, creation_info_dict.get("licenseListVersion"), self.parse_version
        )
        document_comment: Optional[str] = doc_dict.get("comment")
        raise_parsing_error_if_logger_has_messages(logger, "Document")

        creation_info = construct_or_raise_parsing_error(
            CreationInfo,
            dict(
                spdx_version=spdx_version,
                spdx_id=spdx_id,
                name=name,
                document_namespace=document_namespace,
                creators=creators,
                created=created,
                license_list_version=license_list_version,
                document_comment=document_comment,
                creator_comment=creator_comment,
                data_license=data_license,
                external_document_refs=external_document_refs,
            ),
        )

        return creation_info

    @staticmethod
    def parse_version(version_str: str) -> Version:
        try:
            return Version.from_string(version_str)
        except ValueError as err:
            raise SPDXParsingError([f"Error while parsing version {version_str}: {err.args[0]}"])

    def parse_external_document_refs(self, external_document_ref_dicts: List[Dict]) -> List[ExternalDocumentRef]:
        logger = Logger()
        external_document_refs = []
        for external_document_ref_dict in external_document_ref_dicts:
            external_document_ref: ExternalDocumentRef = parse_field_or_log_error(
                logger, external_document_ref_dict, self.parse_external_document_ref
            )

            external_document_refs.append(external_document_ref)

        raise_parsing_error_if_logger_has_messages(logger)
        return external_document_refs

    def parse_external_document_ref(self, external_document_ref_dict: Dict) -> ExternalDocumentRef:
        logger = Logger()
        checksum: Optional[Checksum] = parse_field_or_log_error(
            logger, external_document_ref_dict.get("checksum"), self.checksum_parser.parse_checksum
        )

        external_document_id: Optional[str] = external_document_ref_dict.get("externalDocumentId")
        document_uri: Optional[str] = external_document_ref_dict.get("spdxDocument")
        raise_parsing_error_if_logger_has_messages(logger, "ExternalDocumentRef")
        external_document_ref: ExternalDocumentRef = construct_or_raise_parsing_error(
            ExternalDocumentRef,
            dict(document_ref_id=external_document_id, checksum=checksum, document_uri=document_uri),
        )

        return external_document_ref



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2022 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm


def write_checksum_to_tag_value(checksum: Checksum) -> str:
    algorithm_name: str = checksum.algorithm.name
    # Convert underscores to dashes, and other Blake2b-specific casing rules
    if "_" in algorithm_name:
        algorithm_name = CHECKSUM_ALGORITHM_TO_TV.get(algorithm_name)
        if algorithm_name is None:
            raise ValueError(f"Missing conversion rule for converting {checksum.algorithm.name} to tag-value string")
    return f"{algorithm_name}: {checksum.value}"


CHECKSUM_ALGORITHM_TO_TV = {
    ChecksumAlgorithm.BLAKE2B_256.name: "BLAKE2b-256",
    ChecksumAlgorithm.BLAKE2B_384.name: "BLAKE2b-384",
    ChecksumAlgorithm.BLAKE2B_512.name: "BLAKE2b-512",
    ChecksumAlgorithm.SHA3_256.name: "SHA3-256",
    ChecksumAlgorithm.SHA3_384.name: "SHA3-384",
    ChecksumAlgorithm.SHA3_512.name: "SHA3-512",
}



# SPDX-FileCopyrightText: 2022 spdx contributors

# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.bump_from_spdx2.bump_utils import handle_no_assertion_or_none
from spdx_tools.spdx.model.spdx_no_assertion import SpdxNoAssertion
from spdx_tools.spdx.model.spdx_none import SpdxNone


@pytest.mark.parametrize(
    "input_argument,expected_value,expected_stdout",
    [
        (SpdxNone(), None, "test_field: Missing conversion for SpdxNone.\n"),
        (SpdxNoAssertion(), None, ""),
        ("test_string", "test_string", ""),
    ],
)
def test_handle_no_assertion_or_none(input_argument, expected_value, expected_stdout, capsys):
    value = handle_no_assertion_or_none(input_argument, "test_field")

    captured = capsys.readouterr()

    assert value == expected_value
    assert captured.out == expected_stdout



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict, List, Union

from spdx_tools.spdx.model import File, Package, Snippet

try:
    from networkx import DiGraph
except ImportError:
    DiGraph = None
from spdx_tools.spdx.document_utils import get_contained_spdx_elements
from spdx_tools.spdx.model import Document, Relationship


def export_graph_from_document(document: Document, file_name: str) -> None:
    from networkx.drawing import nx_agraph

    graph = generate_relationship_graph_from_spdx(document)
    _color_nodes(graph)
    attributes_graph = nx_agraph.to_agraph(graph)  # convert to a pygraphviz graph
    attributes_graph.draw(file_name, prog="dot")


def generate_relationship_graph_from_spdx(document: Document) -> DiGraph:
    from networkx import DiGraph

    graph = DiGraph()
    graph.add_node(document.creation_info.spdx_id, element=document.creation_info)

    contained_elements: Dict[str, Union[Package, File, Snippet]] = get_contained_spdx_elements(document)
    contained_element_nodes = [(spdx_id, {"element": element}) for spdx_id, element in contained_elements.items()]
    graph.add_nodes_from(contained_element_nodes)

    relationships_by_spdx_id: Dict[str, List[Relationship]] = dict()
    for relationship in document.relationships:
        relationships_by_spdx_id.setdefault(relationship.spdx_element_id, []).append(relationship)

    for spdx_id, relationships in relationships_by_spdx_id.items():
        if spdx_id not in graph.nodes():
            # this will add any external spdx_id to the graph where we have no further information about the element,
            # to indicate that this node represents an element we add the attribute "element"
            graph.add_node(spdx_id, element=None)
        for relationship in relationships:
            relationship_node_key = relationship.spdx_element_id + "_" + relationship.relationship_type.name
            graph.add_node(relationship_node_key, comment=relationship.comment)
            graph.add_edge(relationship.spdx_element_id, relationship_node_key)
            # if the related spdx element is SpdxNone or SpdxNoAssertion we need a type conversion
            related_spdx_element_id = str(relationship.related_spdx_element_id)

            if related_spdx_element_id not in graph.nodes():
                # this will add any external spdx_id to the graph where we have no further information about
                # the element, to indicate that this node represents an element we add the attribute "element"
                graph.add_node(
                    related_spdx_element_id,
                    element=None,
                )
            graph.add_edge(relationship_node_key, related_spdx_element_id)

    return graph


def _color_nodes(graph: DiGraph) -> None:
    for node in graph.nodes():
        if "_" in node:
            # nodes representing a RelationshipType are concatenated with the spdx_element_id,
            # to only see the RelationshipType when rendering the graph to a picture we add
            # a label to these nodes
            graph.add_node(node, color="lightgreen", label=node.split("_", 1)[-1])
        elif node == "SPDXRef-DOCUMENT":
            graph.add_node(node, color="indianred2")
        else:
            graph.add_node(node, color="lightskyblue")



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from datetime import datetime

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx.model import (
    Actor,
    Annotation,
    ExternalDocumentRef,
    ExtractedLicensingInfo,
    File,
    Package,
    Relationship,
    Snippet,
    Version,
)


@dataclass_with_properties
class CreationInfo:
    spdx_version: str
    spdx_id: str
    name: str
    document_namespace: str
    creators: List[Actor]
    created: datetime
    creator_comment: Optional[str] = None
    data_license: str = "CC0-1.0"
    external_document_refs: List[ExternalDocumentRef] = field(default_factory=list)
    license_list_version: Optional[Version] = None
    document_comment: Optional[str] = None

    def __init__(
        self,
        spdx_version: str,
        spdx_id: str,
        name: str,
        document_namespace: str,
        creators: List[Actor],
        created: datetime,
        creator_comment: Optional[str] = None,
        data_license: str = "CC0-1.0",
        external_document_refs: List[ExternalDocumentRef] = None,
        license_list_version: Optional[Version] = None,
        document_comment: Optional[str] = None,
    ):
        external_document_refs = [] if external_document_refs is None else external_document_refs
        check_types_and_set_values(self, locals())


@dataclass_with_properties
class Document:
    creation_info: CreationInfo

    packages: List[Package] = field(default_factory=list)
    files: List[File] = field(default_factory=list)
    snippets: List[Snippet] = field(default_factory=list)
    annotations: List[Annotation] = field(default_factory=list)
    relationships: List[Relationship] = field(default_factory=list)
    extracted_licensing_info: List[ExtractedLicensingInfo] = field(default_factory=list)

    def __init__(
        self,
        creation_info: CreationInfo,
        packages: List[Package] = None,
        files: List[File] = None,
        snippets: List[Snippet] = None,
        annotations: List[Annotation] = None,
        relationships: List[Relationship] = None,
        extracted_licensing_info: List[ExtractedLicensingInfo] = None,
    ):
        packages = [] if packages is None else packages
        files = [] if files is None else files
        snippets = [] if snippets is None else snippets
        annotations = [] if annotations is None else annotations
        relationships = [] if relationships is None else relationships
        extracted_licensing_info = [] if extracted_licensing_info is None else extracted_licensing_info
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model.software import Sbom
from spdx_tools.spdx3.writer.console.bom_writer import write_bom


def write_sbom(sbom: Sbom, text_output: TextIO):
    text_output.write("## Sbom\n")
    write_bom(sbom, text_output, False)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.model import PackageVerificationCode


def test_correct_initialization():
    package_verification_code = PackageVerificationCode("value", ["file1", "file2"])
    assert package_verification_code.value == "value"
    assert package_verification_code.excluded_files == ["file1", "file2"]


def test_wrong_type_in_value():
    with pytest.raises(TypeError):
        PackageVerificationCode(42, ["file1", "file2"])


def test_wrong_type_in_excluded_files():
    with pytest.raises(TypeError):
        PackageVerificationCode("value", "file1")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import (
    Annotation,
    Bom,
    Bundle,
    Organization,
    Person,
    Relationship,
    SoftwareAgent,
    SpdxDocument,
    Tool,
)
from spdx_tools.spdx3.model.ai import AIPackage
from spdx_tools.spdx3.model.build import Build
from spdx_tools.spdx3.model.dataset import Dataset
from spdx_tools.spdx3.model.software import File, Package, Sbom, Snippet, SoftwareDependencyRelationship
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx3.writer.console.agent_writer import write_agent
from spdx_tools.spdx3.writer.console.ai.ai_package_writer import write_ai_package
from spdx_tools.spdx3.writer.console.annotation_writer import write_annotation
from spdx_tools.spdx3.writer.console.bom_writer import write_bom
from spdx_tools.spdx3.writer.console.build.build_writer import write_build
from spdx_tools.spdx3.writer.console.bundle_writer import write_bundle
from spdx_tools.spdx3.writer.console.dataset.dataset_writer import write_dataset
from spdx_tools.spdx3.writer.console.relationship_writer import write_relationship
from spdx_tools.spdx3.writer.console.software.file_writer import write_file
from spdx_tools.spdx3.writer.console.software.package_writer import write_package
from spdx_tools.spdx3.writer.console.software.sbom_writer import write_sbom
from spdx_tools.spdx3.writer.console.software.snippet_writer import write_snippet
from spdx_tools.spdx3.writer.console.software.software_dependency_relationship_writer import (
    write_software_dependency_relationship,
)
from spdx_tools.spdx3.writer.console.spdx_document_writer import write_spdx_document
from spdx_tools.spdx3.writer.console.tool_writer import write_tool

MAP_CLASS_TO_WRITE_METHOD = {
    Annotation: write_annotation,
    Relationship: write_relationship,
    SoftwareDependencyRelationship: write_software_dependency_relationship,
    Bundle: write_bundle,
    SpdxDocument: write_spdx_document,
    Bom: write_bom,
    File: write_file,
    Package: write_package,
    Snippet: write_snippet,
    Sbom: write_sbom,
    Person: write_agent,
    Organization: write_agent,
    SoftwareAgent: write_agent,
    Tool: write_tool,
    AIPackage: write_ai_package,
    Dataset: write_dataset,
    Build: write_build,
}


def write_payload(payload: Payload, text_output: TextIO):
    for element in payload.get_full_map().values():
        write_method = MAP_CLASS_TO_WRITE_METHOD[type(element)]
        write_method(element, text_output)
        text_output.write("\n")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Optional
from pyshacl import validate
from rdflib import Graph


def validate_against_shacl_from_file(
    data_file: str, shacl_file: str, data_format: Optional[str] = "json-ld", shacl_format: Optional[str] = "ttl"
):
    data_graph = Graph()
    with open(data_file) as file:
        data_graph.parse(file, format=data_format)

    shacl_graph = Graph()
    with open(shacl_file) as file:
        shacl_graph.parse(file, format=shacl_format)

    return validate(data_graph=data_graph, shacl_graph=shacl_graph, ont_graph=shacl_graph)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase

from spdx_tools.spdx3.bump_from_spdx2.checksum import bump_checksum
from spdx_tools.spdx3.bump_from_spdx2.spdx_document import bump_spdx_document
from spdx_tools.spdx3.model import ExternalMap
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model import ExternalDocumentRef
from spdx_tools.spdx.model.document import Document as Spdx2_Document
from tests.spdx.fixtures import (
    checksum_fixture,
    creation_info_fixture,
    document_fixture,
    file_fixture,
    package_fixture,
    snippet_fixture,
)


def test_bump_external_elements():
    external_doc_uri = "https://external-document.uri"
    external_doc_id = "DocumentRef-external"

    full_external_doc_id = external_doc_id + ":SPDXRef-DOCUMENT"
    package_id = external_doc_id + ":SPDXRef-Package"
    file_id = external_doc_id + ":SPDXRef-File"
    snippet_id = external_doc_id + ":SPDXRef-Snippet"
    document_namespace = document_fixture().creation_info.document_namespace

    spdx2_document: Spdx2_Document = document_fixture(
        creation_info=creation_info_fixture(
            external_document_refs=[ExternalDocumentRef(external_doc_id, external_doc_uri, checksum_fixture())]
        ),
        packages=[package_fixture(spdx_id=package_id)],
        files=[file_fixture(spdx_id=file_id)],
        snippets=[snippet_fixture(spdx_id=snippet_id)],
    )
    payload: Payload = bump_spdx_document(spdx2_document)

    expected_imports = [
        ExternalMap(external_id=package_id, defining_document=full_external_doc_id),
        ExternalMap(external_id=file_id, defining_document=full_external_doc_id),
        ExternalMap(external_id=snippet_id, defining_document=full_external_doc_id),
        ExternalMap(external_id=full_external_doc_id, verified_using=[bump_checksum(checksum_fixture())]),
    ]
    spdx_document = payload.get_element(f"{document_namespace}#SPDXRef-DOCUMENT")

    assert f"{external_doc_uri}#SPDXRef-Package" in payload.get_full_map()
    assert f"{external_doc_uri}#SPDXRef-File" in payload.get_full_map()
    assert f"{external_doc_uri}#SPDXRef-Snippet" in payload.get_full_map()

    TestCase().assertCountEqual(spdx_document.imports, expected_imports)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import re

import uritools
from beartype.typing import Dict, List

from spdx_tools.spdx.model import ExternalPackageRef, ExternalPackageRefCategory
from spdx_tools.spdx.model.package import CATEGORY_TO_EXTERNAL_PACKAGE_REF_TYPES
from spdx_tools.spdx.validation.uri_validators import validate_url
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage

CPE22TYPE_REGEX = r"^c[pP][eE]:/[AHOaho]?(:[A-Za-z0-9._\-~%]*){0,6}$"
CPE23TYPE_REGEX = (
    r'^cpe:2\.3:[aho\*\-](:(((\?*|\*?)([a-zA-Z0-9\-\._]|(\\[\\\*\?!"#$$%&\'\(\)\+,\/:;<=>@\[\]\^'
    r"`\{\|}~]))+(\?*|\*?))|[\*\-])){5}(:(([a-zA-Z]{2,3}(-([a-zA-Z]{2}|[0-9]{3}))?)|[\*\-]))(:(((\?*"
    r'|\*?)([a-zA-Z0-9\-\._]|(\\[\\\*\?!"#$$%&\'\(\)\+,\/:;<=>@\[\]\^`\{\|}~]))+(\?*|\*?))|[\*\-])){4}$'
)
MAVEN_CENTRAL_REGEX = r"^[^:]+:[^:]+(:[^:]+)?$"
NPM_REGEX = r"^[^@]+@[^@]+$"
NUGET_REGEX = r"^[^/]+/[^/]+$"
BOWER_REGEX = r"^[^#]+#[^#]+$"
PURL_REGEX = r"^pkg:.+(\/.+)?\/.+(@.+)?(\?.+)?(#.+)?$"
SWH_REGEX = r"^swh:1:(snp|rel|rev|dir|cnt):[0-9a-fA-F]{40}$"
GITOID_REGEX = r"^gitoid:(blob|tree|commit|tag):(sha1:[0-9a-fA-F]{40}|sha256:[0-9a-fA-F]{64})$"

TYPE_TO_REGEX: Dict[str, str] = {
    "cpe22Type": CPE22TYPE_REGEX,
    "cpe23Type": CPE23TYPE_REGEX,
    "maven-central": MAVEN_CENTRAL_REGEX,
    "npm": NPM_REGEX,
    "nuget": NUGET_REGEX,
    "bower": BOWER_REGEX,
    "purl": PURL_REGEX,
    "swh": SWH_REGEX,
    "gitoid": GITOID_REGEX,
}


def validate_external_package_refs(
    external_package_refs: List[ExternalPackageRef], parent_id: str, spdx_version: str
) -> List[ValidationMessage]:
    validation_messages = []
    for external_package_ref in external_package_refs:
        validation_messages.extend(validate_external_package_ref(external_package_ref, parent_id, spdx_version))

    return validation_messages


def validate_external_package_ref(
    external_package_ref: ExternalPackageRef, parent_id: str, spdx_version: str
) -> List[ValidationMessage]:
    validation_messages = []
    context = ValidationContext(
        parent_id=parent_id, element_type=SpdxElementType.EXTERNAL_PACKAGE_REF, full_element=external_package_ref
    )

    category = external_package_ref.category
    locator = external_package_ref.locator
    reference_type = external_package_ref.reference_type

    if category == ExternalPackageRefCategory.OTHER:
        if " " in locator:
            validation_messages.append(
                ValidationMessage(
                    f"externalPackageRef locator in category OTHER must contain no spaces, but is: {locator}", context
                )
            )

    elif spdx_version == "SPDX-2.2" and reference_type in ["advisory", "fix", "url", "swid"]:
        return [ValidationMessage(f'externalPackageRef type "{reference_type}" is not supported in SPDX-2.2', context)]

    elif reference_type not in CATEGORY_TO_EXTERNAL_PACKAGE_REF_TYPES[category]:
        validation_messages.append(
            ValidationMessage(
                f"externalPackageRef type in category {category.name} must be one of "
                f"{CATEGORY_TO_EXTERNAL_PACKAGE_REF_TYPES[category]}, but is: {reference_type}",
                context,
            )
        )

    elif reference_type in ["advisory", "fix", "url"]:
        if validate_url(locator):
            validation_messages.append(
                ValidationMessage(
                    f'externalPackageRef locator of type "{reference_type}" must be a valid URL, but is: {locator}',
                    context,
                )
            )

    elif reference_type == "swid":
        if not uritools.isuri(locator) or not locator.startswith("swid"):
            validation_messages.append(
                ValidationMessage(
                    f'externalPackageRef locator of type "swid" must be a valid URI with scheme swid, '
                    f"but is: {locator}",
                    context,
                )
            )

    else:
        validation_messages.extend(validate_against_regex(locator, reference_type, context))

    return validation_messages


def validate_against_regex(
    string_to_validate: str, reference_type: str, context: ValidationContext
) -> List[ValidationMessage]:
    regex = TYPE_TO_REGEX[reference_type]
    if not re.match(regex, string_to_validate):
        return [
            ValidationMessage(
                f'externalPackageRef locator of type "{reference_type}" must conform with the regex {regex}, '
                f"but is: {string_to_validate}",
                context,
            )
        ]
    return []



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os

import pytest
from rdflib import RDF, Graph

from spdx_tools.spdx.parser.rdf.extracted_licensing_info_parser import parse_extracted_licensing_info
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


@pytest.mark.parametrize(
    "license_id, extracted_text, comment, license_name, cross_references",
    [
        (
            "LicenseRef-1",
            "extractedText",
            "licenseComment",
            "licenseName",
            ["https://see.also"],
        ),
        (
            "LicenseRef-2",
            "extractedText",
            "licenseComment",
            "another license",
            ["https://see.also"],
        ),
    ],
)
# In rdf, as a short form, the explicit node for licenseId can be omitted, since the ID is also encoded in the URIRef
# of the ExtractedLicensingInfo node. The first test case has an explicit licenseId node whereas the second test case
# does not. This behaviour is similar to the externalDocumentRefId, see the discussion here:
# https://github.com/spdx/spdx-spec/issues/816
def test_parse_extracted_licensing_info(license_id, extracted_text, comment, license_name, cross_references):
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    doc_namespace = "https://some.namespace"
    extracted_licensing_info_node = get_extracted_licensing_info_node_by_license_id(graph, license_id)

    extracted_licensing_info = parse_extracted_licensing_info(extracted_licensing_info_node, graph, doc_namespace)

    assert extracted_licensing_info.license_id == license_id
    assert extracted_licensing_info.extracted_text == extracted_text
    assert extracted_licensing_info.comment == comment
    assert extracted_licensing_info.license_name == license_name
    assert extracted_licensing_info.cross_references == cross_references


def get_extracted_licensing_info_node_by_license_id(graph, license_id):
    for extracted_licensing_info_node, _, _ in graph.triples((None, RDF.type, SPDX_NAMESPACE.ExtractedLicensingInfo)):
        if extracted_licensing_info_node.fragment == license_id:
            return extracted_licensing_info_node



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os

import pytest
from click.testing import CliRunner

from spdx_tools.spdx.clitools.pyspdxtools import main


@pytest.mark.parametrize(
    "options",
    [
        ("--infile", os.path.join(os.path.dirname(__file__), "data/SPDXJSONExample-v2.3.spdx.json")),
        ("-i", os.path.join(os.path.dirname(__file__), "data/SPDXJSONExample-v2.3.spdx.json"), "--novalidation"),
        (
            "-i",
            os.path.join(os.path.dirname(__file__), "data/SPDXJSONExample-v2.3.spdx.json"),
            "--novalidation",
            "--version",
            "SPDX-2.3",
        ),
        ("-i", os.path.join(os.path.dirname(__file__), "data/SPDXJSONExample-v2.3.spdx.json"), "-o", "-"),
    ],
)
def test_cli_with_system_exit_code_0(options):
    runner = CliRunner()

    result = runner.invoke(main, options)

    assert result.exit_code == 0


@pytest.mark.parametrize(
    "options",
    [
        (),
        ("-i", os.path.join(os.path.dirname(__file__), "data/SPDXJSONExample-v2.3.spdx.json"), "--version"),
        ("-i", os.path.join(os.path.dirname(__file__), "data/SPDXJSONExample-v2.3.spdx.json"), "-o"),
    ],
)
def test_cli_with_system_exit_code_2(options):
    runner = CliRunner()

    result = runner.invoke(main, options)

    assert result.exit_code == 2



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDF, BNode, Graph, Literal, URIRef

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def add_checksum_to_graph(checksum: Checksum, graph: Graph, parent: URIRef):
    checksum_node = BNode()
    graph.add((checksum_node, RDF.type, SPDX_NAMESPACE.Checksum))
    graph.add((checksum_node, SPDX_NAMESPACE.algorithm, algorithm_to_rdf_string(checksum.algorithm)))
    graph.add((checksum_node, SPDX_NAMESPACE.checksumValue, Literal(checksum.value)))

    graph.add((parent, SPDX_NAMESPACE.checksum, checksum_node))


def algorithm_to_rdf_string(algorithm: ChecksumAlgorithm) -> URIRef:
    if "BLAKE2B" in algorithm.name:
        algorithm_rdf_string = algorithm.name.replace("_", "").lower()
    else:
        algorithm_rdf_string = algorithm.name.lower()

    return SPDX_NAMESPACE[f"checksumAlgorithm_{algorithm_rdf_string}"]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from copy import deepcopy

from spdx_tools.spdx3.bump_from_spdx2.actor import bump_actor
from spdx_tools.spdx3.bump_from_spdx2.message import print_missing_conversion
from spdx_tools.spdx3.model import Annotation, AnnotationType, CreationInfo
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model.actor import ActorType
from spdx_tools.spdx.model.annotation import Annotation as Spdx2_Annotation


def bump_annotation(
    spdx2_annotation: Spdx2_Annotation,
    payload: Payload,
    creation_info: CreationInfo,
    document_namespace: str,
    counter: int,
):
    spdx_id: str = "#".join([document_namespace, f"SPDXRef-Annotation-{counter}"])
    creation_info = deepcopy(creation_info)
    creation_info.created = spdx2_annotation.annotation_date

    # caution: the annotator and the annotation will only share the same creation_info if the actor
    #          has not been previously defined
    annotator = spdx2_annotation.annotator
    creator_id: str = bump_actor(annotator, payload, document_namespace, creation_info)
    if annotator.actor_type in [ActorType.PERSON, ActorType.ORGANIZATION]:
        creation_info.created_by = [creator_id]
    else:
        print_missing_conversion(
            "Annotator",
            0,
            "The SPDX2 annotation is not of Type Person or Organization."
            " This case leads to an invalid SPDX3 document and is currently not supported."
            "https://github.com/spdx/spdx-3-model/issues/180",
        )
    annotation_type: AnnotationType = AnnotationType[spdx2_annotation.annotation_type.name]

    payload.add_element(
        Annotation(
            spdx_id,
            annotation_type,
            creation_info=creation_info,
            subject=f"{document_namespace}#{spdx2_annotation.spdx_id}",
            statement=spdx2_annotation.annotation_comment,
        )
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import ElementCollection
from spdx_tools.spdx3.writer.console.element_writer import write_element_properties
from spdx_tools.spdx3.writer.console.external_map_writer import write_external_map
from spdx_tools.spdx3.writer.console.namespace_map_writer import write_namespace_map
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import write_optional_heading


def write_collection(collection: ElementCollection, text_output: TextIO):
    write_element_properties(collection, text_output)
    text_output.write(f"elements: {', '.join(collection.element)}\n")
    write_optional_heading(collection.namespaces, "# Namespaces\n", text_output)
    for namespace_map in collection.namespaces:
        write_namespace_map(namespace_map, text_output)
    write_optional_heading(collection.imports, "# Imports\n", text_output)
    for external_map in collection.imports:
        write_external_map(external_map, text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Dict

from spdx_tools.common.typing.constructor_type_errors import ConstructorTypeErrors
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.logger import Logger


def construct_or_raise_parsing_error(object_to_construct: Any, args_for_construction: Dict) -> Any:
    try:
        constructed_object = object_to_construct(**args_for_construction)
    except ConstructorTypeErrors as err:
        raise SPDXParsingError([f"Error while constructing {object_to_construct.__name__}: {err.get_messages()}"])
    except TypeError as err:
        raise SPDXParsingError([f"Error while constructing {object_to_construct.__name__}: {err.args[0]}"])
    return constructed_object


def raise_parsing_error_if_logger_has_messages(logger: Logger, parsed_object_name: str = None):
    if logger.has_messages():
        if parsed_object_name:
            raise SPDXParsingError([f"Error while parsing {parsed_object_name}: {logger.get_messages()}"])
        else:
            raise SPDXParsingError(logger.get_messages())
